# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Utility functions """

import json
from warnings import warn

import jsonschema
import pooch
import yaml


def get_catalog_jsonschema(url, known_hash, required):
    """
    Download a jsonschema from a url. Returns the unaltered jsonschema and a version with the "required" key
    matching the properties provided.

    Parameters
    ----------
    url: str
        The URL to the jsonschema file. ACCESS-NRI schema can be found at
        https://github.com/ACCESS-NRI/schema.
    known_hash: str
        A known hash (checksum) of the file. See :py:func:`~pooch.retrieve`.
    required: list
        A list of the properties to include in the "required" key
    """

    schema_file = pooch.retrieve(url=url, known_hash=known_hash)

    with open(schema_file) as fpath:
        schema = json.load(fpath)

    catalog_schema = schema.copy()
    req = []
    for col in required:
        if col not in catalog_schema["properties"]:
            warn(
                f"Required column {col} does not exist in schema. Entries in this column will not be validated"
            )
        else:
            req.append(col)

    catalog_schema["required"] = req

    return schema, catalog_schema


def load_metadata_yaml(path):
    """
    Load a metadata.yaml file, leaving dates as strings and loading arrays as tuples

    Parameters
    ----------
    paths: str
        The path to the metadata.yaml
    """

    class NoDatesSafeLoader(yaml.SafeLoader):
        @classmethod
        def remove_implicit_resolver(cls, tag_to_remove):
            """
            Remove implicit resolvers for a particular tag

            See https://stackoverflow.com/questions/34667108/ignore-dates-and-times-while-parsing-yaml
            """
            if "yaml_implicit_resolvers" not in cls.__dict__:
                cls.yaml_implicit_resolvers = cls.yaml_implicit_resolvers.copy()

            for first_letter, mappings in cls.yaml_implicit_resolvers.items():
                cls.yaml_implicit_resolvers[first_letter] = [
                    (tag, regexp) for tag, regexp in mappings if tag != tag_to_remove
                ]

    def tuple_constructor(self, node):
        """
        yaml constructor to make leaf sequences into tuples

        See https://stackoverflow.com/questions/39553008/how-to-read-a-python-tuple-using-pyyaml
        """
        seq = self.construct_sequence(node)
        if seq and isinstance(seq[0], (list, tuple)):
            return seq
        return tuple(seq)

    NoDatesSafeLoader.remove_implicit_resolver("tag:yaml.org,2002:timestamp")
    NoDatesSafeLoader.add_constructor("tag:yaml.org,2002:seq", tuple_constructor)

    with open(path) as fpath:
        metadata = yaml.load(fpath, Loader=NoDatesSafeLoader)

    return metadata


def validate_against_schema(instance, schema):
    """
    Validate a dictionary against a jsonschema, allowing for tuples as arrays

    Parameters
    ----------
    instance: dict
        The instance to validate
    schema: dict
        The jsonschema
    """

    Validator = jsonschema.validators.validator_for(schema)
    type_checker = Validator.TYPE_CHECKER.redefine(
        "array", lambda checker, instance: isinstance(instance, (list, tuple))
    )
    TupleAllowingValidator = jsonschema.validators.extend(
        Validator, type_checker=type_checker
    )

    TupleAllowingValidator(schema).validate(instance)



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

import argparse
import logging

import jsonschema
import yaml

from .esmcat import builders
from .metacat import METADATA_JSONSCHEMA, manager, translators
from .utils import load_metadata_yaml, validate_against_schema


class MetadataCheckError(Exception):
    pass


def _parse_config_yamls(config_yamls):
    """
    Parse a list of configuration YAML files into a list of tuples of
    MetacatManager methods and args to pass to the methods
    """

    args = []
    for config_yaml in config_yamls:
        with open(config_yaml) as f:
            config = yaml.safe_load(f)

        builder = config.get("builder")
        translator = config.get("translator")
        source_dir = config.get("source_dir")
        sources = config.get("sources")

        config_args = {}
        if builder:
            method = "build_esm"
            config_args["builder"] = getattr(builders, builder)
            config_args["directory"] = source_dir
            config_args["overwrite"] = True
        else:
            method = "load"

        for kwargs in sources:
            source_args = config_args

            source_args["path"] = kwargs.pop("path")
            metadata_yaml = kwargs.pop("metadata_yaml")
            metadata = load_metadata_yaml(metadata_yaml)
            source_args["name"] = metadata["name"]
            source_args["description"] = metadata["description"]
            source_args["metadata"] = metadata

            if translator:
                source_args["translator"] = getattr(translators, translator)

            args.append((method, source_args | kwargs))

    return args


def _check_args(args_list):
    """
    Run some checks on the parsed argmuents to be passed to the MetacatManager
    """

    names = []
    uuids = []
    for args in args_list:
        names.append(args["name"])
        uuids.append(args["metadata"]["experiment_uuid"])
        try:
            validate_against_schema(args["metadata"], METADATA_JSONSCHEMA)
        except jsonschema.exceptions.ValidationError:
            raise MetadataCheckError(
                f"Failed to validate metadata.yaml for {args['name']}. See traceback for details."
            )

    if len(names) != len(set(names)):
        seen = set()
        dupes = [name for name in names if name in seen or seen.add(name)]
        raise MetadataCheckError(f"There are experiments with the same name: {dupes}")
    if len(uuids) != len(set(uuids)):
        seen = set()
        dupes = [uuid for uuid in uuids if uuid in seen or seen.add(uuid)]
        dupes = [name for name, uuid in zip(names, uuids) if uuid in dupes]
        raise MetadataCheckError(
            f"There are experiments with the same experiment_uuid: {dupes}"
        )


def build():
    """
    Build/add intake catalog(s) specified in a YAML configuration file to an intake-dataframe-catalog metacatalog
    """

    log_fmt = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_fmt)
    logger = logging.getLogger(__name__)

    parser = argparse.ArgumentParser(
        description=(
            "Build/add intake catalog(s) specified in a YAML configuration file to an intake-dataframe-catalog "
            "metacatalog"
        )
    )
    parser.add_argument(
        "config_yaml",
        type=str,
        nargs="+",
        help="Configuration YAML file(s) specifying the intake catalog(s) to add",
    )
    parser.add_argument(
        "--catalog_name",
        type=str,
        default="metacatalog.csv",
        help="The path to the intake-dataframe-catalog",
    )

    args = parser.parse_args()
    config_yamls = args.config_yaml
    catalog_name = args.catalog_name

    parsed_sources = _parse_config_yamls(config_yamls)
    _check_args([parsed_source[1] for parsed_source in parsed_sources])

    for (method, args) in parsed_sources:
        man = manager.MetacatManager(path=catalog_name)
        logger.info(f"Adding '{args['name']}' to metacatalog '{catalog_name}'")
        getattr(man, method)(**args).add()



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

"""
Tools for translating metadata in an intake source into a metadata table to use in an intake-dataframe-catalog
"""

import re
from functools import partial

import pandas as pd
import tlz

from . import COLUMNS_WITH_ITERABLES


class TranslatorError(Exception):
    pass


class DefaultTranslator:
    """
    Default Translator for translating metadata in an intake catalog into a :py:class:`~pandas.DataFrame`
    of metadata for use in an intake-dataframe-catalog.
    """

    def __init__(self, cat, columns):
        """
        Initialise a DefaultTranslator. This Translator works as follows:

        - If the input catalog is an intake-esm catalog, the translator will first look for the column in the
             esmcat.df attribute, casting iterable columns to tuples. If the catalog is not an intake-esm catalog,
             this step is skipped.
        - If that fails, the translator will then look for the column name as an attribute on the catalog itself
        - If that fails, the translator will then look for the column name in the metadata attribute of the catalog

        Parameters
        ----------
        cat: :py:class:`~intake.DataSource`
            The catalog to use to do the translations
        columns: list of str
            The columns to translate
        """

        self.cat = cat
        self.columns = columns
        self._dispatch = {
            column: partial(self._default_translator, column=column)
            for column in columns
        }

    def _default_translator(self, column):
        """
        Try to translate a column from a catalog using the default translator. This translator works as follows:
        - If the input catalog is an intake-esm catalog, the translator will first look for the column in the
             esmcat.df attribute, casting iterable columns to tuples. If the catalog is not an intake-esm catalog,
             this step is skipped.
        - If that fails, the translator will then look for the column name as an attribute on the catalog itself
        - If that fails, the translator will then look for the column name in the metadata attribute of the catalog

        Parameters
        ----------
        column: str
            The column to translate, e.g. "frequency"
        """
        if hasattr(self.cat, "esmcat"):
            try:
                series = self.cat.df[column]

                # Cast to tuples
                if column in self.cat.esmcat.columns_with_iterables:
                    return series.apply(tuple)
                elif column in COLUMNS_WITH_ITERABLES:
                    return to_tuple(series)
                else:
                    return series
            except KeyError:
                len_df = len(self.cat.df)
        else:
            len_df = 1

        if hasattr(self.cat, column):
            val = getattr(self.cat, column)
        elif column in self.cat.metadata:
            val = self.cat.metadata[column]
        else:
            raise TranslatorError(
                f"Could not translate '{column}' from {self.cat.name} using {self.__class__.__name__}"
            )

        return pd.Series([val] * len_df)

    def translate(self, groupby):
        """
        Return the translated :py:class:`~pandas.DataFrame` of metadata and merge into set of
        set of rows with unique values of the columns specified.

        Parameters
        ----------
        groupby: list of str
            Core metadata columns to group by before merging metadata across remaining core columns.
        """

        def _find_unique(series):
            """
            Return a set of unique values in a series
            """

            values = series.dropna()
            iterable_entries = series.name in COLUMNS_WITH_ITERABLES
            type_ = type(series.iloc[0])

            if iterable_entries:
                values = tlz.concat(values)

            uniques = tuple(set(values))

            return (
                uniques[0]
                if (len(uniques) == 1) & (not iterable_entries)
                else type_(uniques)
            )

        df = pd.concat(
            {col: func() for col, func in self._dispatch.items()}, axis="columns"
        )

        ungrouped_columns = list(set(self.columns) - set(groupby))
        df_grouped = (
            df.groupby(groupby)
            .agg({col: _find_unique for col in ungrouped_columns})
            .reset_index()
        )

        return df_grouped[self.columns]  # Preserve ordering


class Cmip6Translator(DefaultTranslator):
    """
    CMIP6 Translator for translating metadata from the NCI CMIP6 intake catalogs.
    """

    def __init__(self, cat, columns):
        """
        Initialise a Cmip6Translator

        Parameters
        ----------
        cat: :py:class:`~intake.DataSource`
            The NCI CMIP6 intake-esm catalog
        columns: list of str
            The columns to translate
        """

        super().__init__(cat, columns)
        self._dispatch["model"] = self._model_translator
        self._dispatch["realm"] = self._realm_translator
        self._dispatch["frequency"] = self._frequency_translator
        self._dispatch["variable"] = self._variable_translator

    def _model_translator(self):
        """
        Return model from source_id
        """
        return to_tuple(self.cat.df["source_id"])

    def _realm_translator(self):
        """
        Return realm, fixing a few issues
        """
        return _cmip_realm_translator(self.cat.df)

    def _frequency_translator(self):
        """
        Return frequency, fixing a few issues
        """
        return _cmip_frequency_translator(self.cat.df)

    def _variable_translator(self):
        """
        Return variable as a tuple
        """
        return to_tuple(self.cat.df["variable_id"])


class Cmip5Translator(DefaultTranslator):
    """
    CMIP5 Translator for translating metadata from the NCI CMIP5 intake catalogs.
    """

    def __init__(self, cat, columns):
        """
        Initialise a Cmip5Translator

        Parameters
        ----------
        cat: :py:class:`~intake.DataSource`
            The NCI CMIP5 intake-esm catalog
        columns: list of str
            The columns to translate
        """

        super().__init__(cat, columns)
        self._dispatch["model"] = self._model_translator
        self._dispatch["realm"] = self._realm_translator
        self._dispatch["frequency"] = self._frequency_translator
        self._dispatch["variable"] = self._variable_translator

    def _model_translator(self):
        """
        Return variable as a tuple
        """
        return to_tuple(self.cat.df["model"])

    def _realm_translator(self):
        """
        Return realm, fixing a few issues
        """
        return _cmip_realm_translator(self.cat.df)

    def _frequency_translator(self):
        """
        Return frequency, fixing a few issues
        """
        return _cmip_frequency_translator(self.cat.df)

    def _variable_translator(self):
        """
        Return variable as a tuple
        """
        return to_tuple(self.cat.df["variable"])


class EraiTranslator(DefaultTranslator):
    """
    ERAI Translator for translating metadata from the NCI ERA-Interim intake catalogs.
    """

    def __init__(self, cat, columns):
        """
        Initialise a EraiTranslator

        Parameters
        ----------
        cat: :py:class:`~intake.DataSource`
            The NCI ERA-Interim intake-esm catalog
        columns: list of str
            The columns to translate
        """

        super().__init__(cat, columns)
        self._dispatch["variable"] = self._variable_translator

    def _variable_translator(self):
        """
        Return variable as a tuple
        """
        return to_tuple(self.cat.df["variable"])


def _cmip_frequency_translator(df):
    """
    Return frequency from CMIP frequency metadata
    """

    def _parse(string):
        for remove in ["Pt", "C.*"]:  # Remove Pt, C, and Clim
            string = re.sub(remove, "", string)
        string = string.replace("daily", "day")  # Some incorrect metadata
        string = string.replace("sem", "3mon")  # CORDEX for seasonal mean
        return (f"1{string}",) if string[0] in ["m", "d", "y"] else (string,)

    return df["frequency"].apply(lambda string: _parse(string))


def _cmip_realm_translator(df):
    """
    Return realm from CMIP realm metadata, fixing some issues
    """

    def _parse(string):
        raw_realms = string.split(" ")
        realms = []
        for realm in raw_realms:
            if re.match("na", realm, flags=re.I):
                realms.append("none")
            elif re.match("seaIce", realm, flags=re.I):
                realms.append("seaIce")
            elif re.match("landIce", realm, flags=re.I):
                realms.append("landIce")
            elif re.match("ocnBgchem", realm, flags=re.I):
                realms.append("ocnBgchem")
            elif re.match("atmos", realm, flags=re.I):
                realms.append("atmos")
            elif re.match("atmosChem", realm, flags=re.I):
                realms.append("atmosChem")
            elif re.match("aerosol", realm, flags=re.I):
                realms.append("aerosol")
            elif re.match("land", realm, flags=re.I):
                realms.append("land")
            elif re.match("ocean", realm, flags=re.I):
                realms.append("ocean")
            else:
                realms.append("unknown")
        return tuple(set(realms))

    return df["realm"].apply(lambda string: _parse(string))


def to_tuple(series):
    """
    Make entries in the provided series a tuple

    Parameters
    ----------
    series: :py:class:`~pandas.Series`
        A pandas Series or another object with an `apply` method
    """
    return series.apply(lambda x: (x,))


def match_substrings(series, substrings):
    """
    Search for a list of substrings in each entry, ignoring case, and return the
    one that's found.

    Parameters
    ----------
    series: :py:class:`~pandas.Series`
        A pandas Series or another object with an `apply` method
    substrings: list of str
        A list of substrings to try and match on each entry in series
    """

    def _parse(s):
        for substring in substrings:
            match = re.match(rf".*{substring}.*", s, flags=re.IGNORECASE)
            if match:
                return substring
        raise ValueError(f"Could not match {s} to any substring")

    return series.apply(lambda s: _parse(s))



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Tools for generating intake-esm catalogs """


from ..utils import get_catalog_jsonschema

CORE_COLUMNS = [
    "path",
    "realm",
    "frequency",
    "variable",
    "start_date",
    "end_date",
]
PATH_COLUMN = "path"
VARIABLE_COLUMN = "variable"

SCHEMA_URL = "https://raw.githubusercontent.com/ACCESS-NRI/schema/4e3d10e563d7c1c9f66e9ab92a2926cdec3d6893/file_asset.json"
SCHEMA_HASH = "2a09030653f495939c90a22e95dd1c4587c8695f7f07e17b9129a6491469f9fc"

_, CATALOG_JSONSCHEMA = get_catalog_jsonschema(
    url=SCHEMA_URL, known_hash=SCHEMA_HASH, required=CORE_COLUMNS
)



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Shared utilities for writing intake-esm builders and their parsers """

import re

import cftime


def get_timeinfo(ds, time_dim="time"):
    """
    Get start time, end time and frequency of a xarray dataset. Stolen and slightly adapted
    from cosima cookbook, see
    https://github.com/COSIMA/cosima-cookbook/blob/master/cosima_cookbook/database.py#L565

    Parameters
    ----------
    ds: :py:class:`xarray.Dataset`
        The dataset to parse the time info from
    time_dim: str
        The name of the time dimension
    """

    if time_dim is None:
        return None

    time_var = ds[time_dim]
    has_bounds = hasattr(time_var, "bounds") and time_var.bounds in ds.variables

    def _todate(t):
        return cftime.num2date(t, time_var.units, calendar=time_var.calendar)

    if has_bounds:
        bounds_var = ds.variables[time_var.bounds]
        start_time = _todate(bounds_var[0, 0])
        end_time = _todate(bounds_var[-1, 1])
    else:
        start_time = _todate(time_var[0])
        end_time = _todate(time_var[-1])

    if len(time_var) > 1 or has_bounds:
        if has_bounds:
            next_time = _todate(bounds_var[0, 1])
        else:
            next_time = _todate(time_var[1])

        dt = next_time - start_time
        # TODO: This is not a very good way to get the frequency
        if dt.days >= 365:
            years = round(dt.days / 365)
            frequency = f"{years}yr"
        elif dt.days >= 28:
            months = round(dt.days / 30)
            frequency = f"{months}mon"
        elif dt.days >= 1:
            frequency = f"{dt.days}day"
        else:
            frequency = f"{dt.seconds // 3600}hr"
    else:
        # single time value in this file and no averaging
        frequency = "fx"

    return (
        start_time.strftime("%Y-%m-%d, %H:%M:%S"),
        end_time.strftime("%Y-%m-%d, %H:%M:%S"),
        frequency,
    )


def strip_pattern_rh(patterns, string):
    """
    Sequentially strip a list of regex patterns from a string, starting from the right
    hand side. Then replace any "-" and "." with "_" and then remove double or dangling "_".

    Parameters:
    -----------
    patterns: list of str
        The list of regex patterns to strip
    string: str
        A filename with the suffixe (e.g. .nc) removed
    """

    # Strip first matched pattern
    stripped = string
    for pattern in patterns:
        match = re.match(rf"^.*({pattern}(?!.*{pattern})).*$", stripped)
        if match:
            stripped = stripped[: match.start(1)] + stripped[match.end(1) :]
            break

    # Enforce Python characters
    stripped = re.sub(r"[-.]", "_", stripped)

    # Remove any double or dangling _
    return re.sub(r"__", "_", stripped).strip("_")



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Tools for managing intake-dataframe-catalog metacatalogs """


from ..utils import get_catalog_jsonschema

CORE_COLUMNS = [
    "name",
    "model",
    "description",
    "realm",
    "frequency",
    "variable",
]
YAML_COLUMN = "yaml"
NAME_COLUMN = "name"
TRANSLATOR_GROUPBY_COLUMNS = ["model", "realm", "frequency"]

SCHEMA_URL = "https://raw.githubusercontent.com/ACCESS-NRI/schema/4e3d10e563d7c1c9f66e9ab92a2926cdec3d6893/experiment_asset.json"
SCHEMA_HASH = "b18cf5bdd06a6f5bcdc71dfc80f7336c63eb49f6d6f75c2cd3371e59eee5488b"

METADATA_JSONSCHEMA, CATALOG_JSONSCHEMA = get_catalog_jsonschema(
    url=SCHEMA_URL, known_hash=SCHEMA_HASH, required=CORE_COLUMNS
)

COLUMNS_WITH_ITERABLES = [
    col
    for col in CORE_COLUMNS
    if CATALOG_JSONSCHEMA["properties"][col]["type"] == "array"
]



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Manager for adding/updating intake catalogs in an intake-dataframe-catalog (metacatalogs) """

import os

import intake
from intake_dataframe_catalog.core import DfFileCatalog

from ..utils import validate_against_schema
from . import (
    CATALOG_JSONSCHEMA,
    COLUMNS_WITH_ITERABLES,
    CORE_COLUMNS,
    NAME_COLUMN,
    TRANSLATOR_GROUPBY_COLUMNS,
    YAML_COLUMN,
)
from .translators import DefaultTranslator


class CatalogExistsError(Exception):
    "Exception for trying to write catalog that already exists"
    pass


class MetacatManager:
    """
    Add/update intake sources in an intake-dataframe-catalog (metacatalog)
    """

    def __init__(self, path):
        """
        Initialise a CatalogManager instance to add/update intake sources in a
        intake-dataframe-catalog metacatalog

        Parameters
        ----------
        path: str
            The path to the intake-dataframe-catalog
        """

        self.path = path

        mode = "a" if os.path.exists(path) else "w"

        self.dfcat = DfFileCatalog(
            path=self.path,
            yaml_column=YAML_COLUMN,
            name_column=NAME_COLUMN,
            mode=mode,
            columns_with_iterables=COLUMNS_WITH_ITERABLES,
        )

        self.source = None
        self.source_metadata = None

    def build_esm(
        self,
        name,
        description,
        builder,
        path,
        translator=DefaultTranslator,
        metadata=None,
        directory=None,
        overwrite=False,
        **kwargs,
    ):
        """
        Build an intake-esm catalog

        Parameters
        ----------
        name: str
            The name of the catalog
        description: str
            Description of the contents of the catalog
        builder: subclass of :py:class:`access_nri_catalog.esmcat.BaseBuilder`
            The builder to use to build the intake-esm catalog
        path: str or list of str
            Path or list of paths to crawl for assets/files to add to the catalog.
        translator: :py:class:`~access_nri_catalog.metacat.translators.DefaultTranslator`
            An instance of the :py:class:`~access_nri_catalog.metacat.translators.DefaultTranslator` class
            for translating info in the intake-esm catalog into intake-dataframe-catalog column metadata.
            Defaults to access_nri_catalog.metacat.translators.DefaultTranslator.
        metadata: dict, optional
            Additional info to store in the intake cat.metadata attribute. This info will be available
            to the translator and to users of the catalog
        directory: str
            The directory to save the catalog to. If None, use the current directory
        overwrite: bool, optional
            Whether to overwrite any existing catalog(s) with the same name
        kwargs: dict
            Additional kwargs to pass to the builder
        """

        metadata = metadata or {}

        json_file = os.path.abspath(f"{os.path.join(directory, name)}.json")
        if os.path.isfile(json_file):
            if not overwrite:
                raise CatalogExistsError(
                    f"A catalog already exists for {name}. To overwrite, "
                    "pass `overwrite=True` to CatalogBuilder.build"
                )

        builder = builder(path, **kwargs).build()
        builder.save(name=name, description=description, directory=directory)

        self.source, self.source_metadata = _open_and_translate(
            json_file,
            name,
            description,
            metadata,
            translator,
            columns_with_iterables=list(builder.columns_with_iterables),
        )

        return self

    def load(
        self,
        name,
        description,
        path,
        translator,
        metadata=None,
        **kwargs,
    ):
        """
        Load an existing intake catalog and add it to the metacatalog

        Parameters
        ----------
        name: str
            The name of the catalog
        description: str
            Description of the contents of the catalog
        path: str
            The path to the intake-esm catalog JSON file
        translator: :py:class:`~access_nri_catalog.metacat.translators.DefaultTranslator`
            An instance of the :py:class:`~access_nri_catalog.metacat.translators.DefaultTranslator` class for
            translating intake-esm column metadata into intake-dataframe-catalog column metadata
        metadata: dict, optional
            Additional info to store in the intake cat.metadata attribute. This info will be available to
            the translator and to users of the catalog
        kwargs: dict, optional
            Additional kwargs to pass to :py:class:`~intake.open_esm_datastore`
        """

        if isinstance(path, list):
            if len(path) != 1:
                raise ValueError(
                    "Only a single JSON file can be passed to CatalogManager.load_esm. Received {len(path)}"
                )
            path = path[0]

        metadata = metadata or {}

        self.source, self.source_metadata = _open_and_translate(
            path, name, description, metadata, translator, **kwargs
        )

        return self

    def add(self, **kwargs):
        """
        Add the catalog to an intake-dataframe-catalog

        Parameters
        ----------
        kwargs: dict, optional
            Additional keyword arguments passed to :py:func:`~pandas.DataFrame.to_csv`.
        """

        # Overwrite the catalog name with the name_column entry in metadata
        name = self.source_metadata[NAME_COLUMN].unique()
        if len(name) != 1:
            raise ValueError(
                f"Metadata column '{NAME_COLUMN}' must be the same for all rows "
                "since this corresponds to the catalog name"
            )
        name = name[0]
        self.source.name = name

        # Validate df_metadata against schema
        for idx, row in self.source_metadata.iterrows():
            validate_against_schema(row.to_dict(), CATALOG_JSONSCHEMA)

        overwrite = True
        for _, row in self.source_metadata.iterrows():
            self.dfcat.add(self.source, row.to_dict(), overwrite=overwrite)
            overwrite = False

        self.dfcat.save(**kwargs)


def _open_and_translate(json_file, name, description, metadata, translator, **kwargs):
    """
    Open an esm-datastore, assign name, description and metadata attrs and
    translate using the provided translator
    """
    cat = intake.open_esm_datastore(json_file, **kwargs)
    cat.name = name
    cat.description = description
    cat.metadata = metadata

    metadata = translator(cat, CORE_COLUMNS).translate(TRANSLATOR_GROUPBY_COLUMNS)

    return cat, metadata



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

import os

import intake

from . import _version

__version__ = _version.get_versions()["version"]

_here = os.path.abspath(os.path.dirname(__file__))
cat = intake.open_catalog(os.path.join(_here, "metacatalog.yaml"))
data = cat.access_nri()



# Copyright 2023 ACCESS-NRI and contributors. See the top-level COPYRIGHT file for details.
# SPDX-License-Identifier: Apache-2.0

""" Builders for generating intake-esm catalogs """

import multiprocessing
import re
import traceback
from pathlib import Path

import xarray as xr
from ecgtools.builder import INVALID_ASSET, TRACEBACK, Builder

from ..utils import validate_against_schema
from . import CATALOG_JSONSCHEMA, PATH_COLUMN, VARIABLE_COLUMN
from .utils import get_timeinfo, strip_pattern_rh


class ParserError(Exception):
    pass


class BaseBuilder(Builder):
    """
    Base class for creating intake-esm catalog builders. Not intended for direct use.
    This builds on the ecgtools.Builder class.
    """

    def __init__(
        self,
        path,
        depth=0,
        exclude_patterns=None,
        include_patterns=None,
        data_format="netcdf",
        groupby_attrs=None,
        aggregations=None,
        storage_options=None,
        joblib_parallel_kwargs={"n_jobs": multiprocessing.cpu_count()},
    ):
        """
        This method should be overwritten. The expection is that some of these arguments
        will be hardcoded in sub classes of this class.

        Parameters
        ----------
        path: str or list of str
            Path or list of path to crawl for assets/files.
        depth: int, optional
            Maximum depth to crawl for assets. Default is 0.
        exclude_patterns: list of str, optional
            List of glob patterns to exclude from crawling.
        include_patterns: list of str, optional
            List of glob patterns to include from crawling.
        data_format: str
            The data format. Valid values are 'netcdf', 'reference' and 'zarr'.
        groupby_attrs: List[str]
            Column names (attributes) that define data sets that can be aggegrated.
        aggregations: List[dict]
            List of aggregations to apply to query results, default None
        storage_options: dict, optional
            Parameters passed to the backend file-system such as Google Cloud Storage,
            Amazon Web Service S3
        joblib_parallel_kwargs: dict, optional
            Parameters passed to joblib.Parallel. Default is {}.
        """

        if isinstance(path, str):
            path = [path]

        self.paths = path
        self.depth = depth
        self.exclude_patterns = exclude_patterns
        self.include_patterns = include_patterns
        self.data_format = data_format
        self.groupby_attrs = groupby_attrs
        self.aggregations = aggregations
        self.storage_options = storage_options
        self.joblib_parallel_kwargs = joblib_parallel_kwargs

        super().__post_init_post_parse__()

    def _parse(self):
        super().parse(parsing_func=self.parser)

    def parse(self):
        """
        Parse metadata from assets.
        """
        self._parse()
        return self

    def _save(self, name, description, directory):
        super().save(
            name=name,
            path_column_name=PATH_COLUMN,
            variable_column_name=VARIABLE_COLUMN,
            data_format=self.data_format,
            groupby_attrs=self.groupby_attrs,
            aggregations=self.aggregations,
            esmcat_version="0.0.1",
            description=description,
            directory=directory,
            catalog_type="file",
            to_csv_kwargs={"compression": "gzip"},
        )

    def save(self, name, description, directory=None):
        """
        Save catalog contents to a file.

        Parameters
        ----------
        name: str
            The name of the file to save the catalog to.
        description : str
            Detailed multi-line description of the collection.
        directory: str, optional
            The directory to save the catalog to. If None, use the current directory.
        """

        if self.df.empty:
            raise ValueError(
                "intake-esm catalog has not yet been built. Please run `.build()` first"
            )

        self._save(name, description, directory)

    def validate_parser(self):
        """
        Run the parser on a single file and check the schema of the info being parsed
        """

        if not self.assets:
            raise ValueError(
                "asset list provided is None. Please run `.get_assets()` first"
            )

        for asset in self.assets:
            info = self.parser(asset)
            if INVALID_ASSET not in info:
                validate_against_schema(info, CATALOG_JSONSCHEMA)
                return self

        raise ParserError(
            "Parser returns no valid assets. Try parsing a single file with Builder.parser(file)"
        )

    def build(self):
        """
        Builds a catalog from a list of netCDF files or zarr stores.
        """

        self.get_assets().validate_parser().parse().clean_dataframe()

        return self

    @property
    def columns_with_iterables(self):
        """
        Return a set of the columns that have iterables
        """
        # Stolen from intake-esm.cat.ESMCatalogModel
        if self.df.empty:
            return set()
        has_iterables = (
            self.df.sample(20, replace=True)
            .applymap(type)
            .isin([list, tuple, set])
            .any()
            .to_dict()
        )
        return {column for column, check in has_iterables.items() if check}

    @staticmethod
    def parser(file):
        """
        Parse info from a file asset

        Parameters
        ----------
        file: str
            The path to the file
        """
        # This method should be overwritten
        pass


class AccessOm2Builder(BaseBuilder):
    """Intake-esm catalog builder for ACCESS-OM2 COSIMA datasets"""

    def __init__(self, path):
        """
        Initialise a AccessOm2Builder

        Parameters
        ----------
        path : str or list of str
            Path or list of paths to crawl for assets/files.
        """

        kwargs = dict(
            path=path,
            depth=3,
            exclude_patterns=["*restart*", "*o2i.nc"],
            include_patterns=["*.nc"],
            data_format="netcdf",
            groupby_attrs=["file_id", "frequency"],
            aggregations=[
                {
                    "type": "join_existing",
                    "attribute_name": "start_date",
                    "options": {
                        "dim": "time",
                        "combine": "by_coords",
                    },
                },
            ],
        )

        super().__init__(**kwargs)

    @staticmethod
    def parser(file):
        try:
            match_groups = re.match(
                r".*/([^/]*)/([^/]*)/output\d+/([^/]*)/.*\.nc", file
            ).groups()
            # configuration = match_groups[0]
            # exp_id = match_groups[1]
            realm = match_groups[2]

            if realm == "ice":
                realm = "seaIce"

            filename = Path(file).stem

            # Get file id without any dates
            # - ocean-3d-v-1-monthly-pow02-ym_1958_04.nc
            # - iceh.057-daily.nc
            # - oceanbgc-3d-caco3-1-yearly-mean-y_2015.nc
            file_id = strip_pattern_rh(
                [
                    r"\d{4}[-_]\d{2}[-_]\d{2}",
                    r"\d{4}[-_]\d{2}",
                    r"\d{4}",
                    r"\d{3}",
                    r"\d{2}",
                ],
                filename,
            )

            with xr.open_dataset(
                file,
                chunks={},
                decode_cf=False,
                decode_times=False,
                decode_coords=False,
            ) as ds:
                variable_list = []
                variable_long_name_list = []
                variable_standard_name_list = []
                variable_cell_methods_list = []
                for var in ds.data_vars:
                    attrs = ds[var].attrs
                    if "long_name" in attrs:
                        variable_list.append(var)
                        variable_long_name_list.append(attrs["long_name"])
                    if "standard_name" in attrs:
                        variable_standard_name_list.append(attrs["standard_name"])
                    if "cell_methods" in attrs:
                        variable_cell_methods_list.append(attrs["cell_methods"])

                start_date, end_date, frequency = get_timeinfo(ds)

            info = {
                "path": str(file),
                "realm": realm,
                "variable": variable_list,
                "frequency": frequency,
                "start_date": start_date,
                "end_date": end_date,
                "variable_long_name": variable_long_name_list,
                "variable_standard_name": variable_standard_name_list,
                "variable_cell_methods": variable_cell_methods_list,
                "filename": filename,
                "file_id": file_id,
            }

            return info

        except Exception:
            return {INVALID_ASSET: file, TRACEBACK: traceback.format_exc()}


class AccessEsm15Builder(BaseBuilder):
    """Intake-esm catalog builder for ACCESS-ESM1.5 datasets"""

    def __init__(self, path, ensemble=False):
        """
        Initialise a AccessEsm15Builder

        Parameters
        ----------
        path: str or list of str
            Path or list of paths to crawl for assets/files.
        ensemble: boolean, optional
            Whether to treat each path as a separate member of an ensemble to join
            along a new member dimension
        """

        kwargs = dict(
            path=path,
            depth=3,
            exclude_patterns=["*restart*"],
            include_patterns=["*.nc*"],
            data_format="netcdf",
            groupby_attrs=["file_id", "frequency"],
            aggregations=[
                {
                    "type": "join_existing",
                    "attribute_name": "start_date",
                    "options": {
                        "dim": "time",
                        "combine": "by_coords",
                    },
                },
            ],
        )

        if ensemble:
            kwargs["aggregations"] += [
                {
                    "type": "join_new",
                    "attribute_name": "member",
                },
            ]

        super().__init__(**kwargs)

    @staticmethod
    def parser(file):
        try:
            match_groups = re.match(r".*/([^/]*)/history/([^/]*)/.*\.nc", file).groups()
            exp_id = match_groups[0]
            realm = match_groups[1]

            realm_mapping = {"atm": "atmos", "ocn": "ocean", "ice": "seaIce"}
            realm = realm_mapping[realm]

            filename = Path(file).stem

            # Get file id without any dates or exp_id
            # - iceh_m.2014-06.nc
            # - bz687a.pm107912_mon.nc
            # - bz687a.p7107912_mon.nc
            # - PI-GWL-B2035.pe-109904_dai.nc
            # - PI-1pct-02.pe-011802_dai.nc_dai.nc
            # - ocean_daily.nc-02531231
            file_id = strip_pattern_rh(
                [r"\d{4}[-_]\d{2}", r"\d{8}", r"\d{6}"], filename
            )
            file_id = strip_pattern_rh([exp_id], file_id)

            with xr.open_dataset(
                file,
                chunks={},
                decode_cf=False,
                decode_times=False,
                decode_coords=False,
            ) as ds:
                variable_list = []
                variable_long_name_list = []
                variable_standard_name_list = []
                variable_cell_methods_list = []
                for var in ds.data_vars:
                    attrs = ds[var].attrs
                    if "long_name" in attrs:
                        variable_list.append(var)
                        variable_long_name_list.append(attrs["long_name"])
                    if "standard_name" in attrs:
                        variable_standard_name_list.append(attrs["standard_name"])
                    if "cell_methods" in attrs:
                        variable_cell_methods_list.append(attrs["cell_methods"])

                start_date, end_date, frequency = get_timeinfo(ds)

            info = {
                "path": str(file),
                "realm": realm,
                "variable": variable_list,
                "frequency": frequency,
                "start_date": start_date,
                "end_date": end_date,
                "member": exp_id,
                "variable_long_name": variable_long_name_list,
                "variable_standard_name": variable_standard_name_list,
                "variable_cell_methods": variable_cell_methods_list,
                "filename": filename,
                "file_id": file_id,
            }

            return info

        except Exception:
            return {INVALID_ASSET: file, TRACEBACK: traceback.format_exc()}


# Include this so it is in the documentation
class AccessCm2Builder(AccessEsm15Builder):
    """Intake-esm catalog builder for ACCESS-CM2 datasets"""

    pass
